from pylel.token import Symbols, Token
from .language_core import CORE
from .language_core.lel_call_function import lel_call_function
from .language_core.standard_language_functions import STANDARD
from .find_base_path import find_base_path

def find_in_scope(scope, name):
	if not scope:
		return None
	if name in scope.variables:
		return scope.variables[name]
	return find_in_scope(scope.upper_scope, name)

def _evaluate_list(scope, expr_list):
	return [evaluate_expr(scope, sub_expr) for sub_expr in expr_list]

def evaluate_expr(scope, expr):
	if isinstance(expr, list):
		# Evaluate empty block
		if not expr:
			return Token(Symbols.LIST, [])

		# List of expressions?
		# Evaluate a block in series
		if isinstance(expr[0], list):
			return _evaluate_list(scope, expr)[-1]

		# The rest of the expressions are based on identifiers
		identifier_token = expr[0]
		if identifier_token.type == Symbols.IDENTIFIER:
			# Core language functions
			if identifier_token.value in CORE:
				return CORE[identifier_token.value](evaluate_expr, scope, expr)

			# Standard languages functions that manipulate primitives
			if identifier_token.value in STANDARD:
				evaluated_expr = _evaluate_list(scope, expr[1:])
				return STANDARD[identifier_token.value](*evaluated_expr)

			# Run a scoped function if one is found
			scoped_function = find_in_scope(scope, identifier_token.value)
			if scoped_function and \
				scoped_function.type == Symbols.FUNCTION_REFERENCE:
				evaluated_expr = _evaluate_list(scope, expr[1:])
				return lel_call_function(evaluate_expr, scope, \
						evaluated_expr, scoped_function.value)

		# Try and evaluate them single expression
		return _evaluate_list(scope, expr)[-1]
	else:
		# Return the value of primitives directly in their tokenised form
		if expr.is_token and \
			( \
				expr.type == Symbols.RANGE \
				or expr.type == Symbols.STRING \
				or expr.type == Symbols.NUMBER \
				or expr.type == Symbols.BOOLEAN \
				or expr.type == Symbols.FUNCTION_REFERENCE \
				or expr.type == Symbols.LIST \
			):
			return expr

		# Identifiers will be a function reference or a variable
		if expr.type == Symbols.IDENTIFIER:
			identifier_type = expr.value

			# Pass back variable value. Explicitly check null instead of other
			# falsey values that might really be contained in the variable
			variable_in_scope = find_in_scope(scope, identifier_type)
			if variable_in_scope:
				return variable_in_scope

	raise Exception("Unrecognised expression: {}".format(expr))



from pylel.token import Symbols, Token

EMPTY_LIST = Token(Symbols.LIST, [])

def _join(lel_list):
	if lel_list.type != Symbols.LIST:
		raise Exception("Can't join non-lists. Got {}".format(lel_list.type))
	return Token(Symbols.STRING, "".join([item.value for item in lel_list.value]))

def _length(lel_list):
	if lel_list.type != Symbols.LIST:
		raise Exception("length operates on LIST type. Got {}".format(lel_list.type))
	return Token(Symbols.NUMBER, len(lel_list.value))

def _head(lel_list):
	if lel_list.type != Symbols.LIST:
		raise Exception("head operates on LIST type. Got {}".format(lel_list.type))
	return lel_list.value[0] if lel_list.value else EMPTY_LIST

def _tail(lel_list):
	if lel_list.type != Symbols.LIST:
		raise Exception("tail operates on LIST type. Got {}".format(lel_list.type))
	return Token(Symbols.LIST, lel_list.value[1:])

def _nth(lel_list, nth_number):
	if lel_list.type != Symbols.LIST:
		raise Exception("nth operates on LIST type. Got {}".format(lel_list.type))
	nth = int(nth_number.value)
	if nth >= 0 and nth < len(lel_list.value):
		return lel_list.value[nth]
	raise Exception("nth: bad index {}. Given list has {} elements"\
		.format(nth_number.value, len(lel_list.value)))

def _sublist(lel_list, start, end):
	if lel_list.type != Symbols.LIST:
		raise Exception("sublist operates on LIST type. Got {}".format(lel_list.type))
	list_length = len(lel_list.value)
	start = int(start.value)
	end = int(end.value)

	if start > end:
		raise Exception("start index cannot be greater than the end index for a sublist")
	if start < 0 or end >= list_length:
		raise Exception("sublist indexes out of range. Got start ({}) end ({}) for list of length {}"\
			.format(start, end, list_length))
	return Token(Symbols.LIST, lel_list.value[start:end + 1])

LEL_LIST = {
	"join": _join,
	"length": _length,
	"head": _head,
	"tail": _tail,
	"nth": _nth,
	"sublist": _sublist
}



from pylel.token import Symbols, Token

def _and(x_number, y_number):
	if x_number.type != Symbols.NUMBER or y_number.type != Symbols.NUMBER:
		raise Exception("- only operates on NUMBER type")
	return Token(Symbols.NUMBER, int(x_number.value) & int(y_number.value))

def _or(x_number, y_number):
	if x_number.type != Symbols.NUMBER or y_number.type != Symbols.NUMBER:
		raise Exception("- only operates on NUMBER type")
	return Token(Symbols.NUMBER, int(x_number.value) | int(y_number.value))

def _xor(x_number, y_number):
	if x_number.type != Symbols.NUMBER or y_number.type != Symbols.NUMBER:
		raise Exception("- only operates on NUMBER type")
	return Token(Symbols.NUMBER, int(x_number.value) ^ int(y_number.value))

def _not(x_number):
	if x_number.type != Symbols.NUMBER:
		raise Exception("- only operates on NUMBER type")
	return Token(Symbols.NUMBER, ~int(x_number.value))

def _left_shift(x_number, y_number):
	if x_number.type != Symbols.NUMBER or y_number.type != Symbols.NUMBER:
		raise Exception("- only operates on NUMBER type")
	return Token(Symbols.NUMBER, int(x_number.value) << int(y_number.value))

def _right_shift(x_number, y_number):
	if x_number.type != Symbols.NUMBER or y_number.type != Symbols.NUMBER:
		raise Exception("- only operates on NUMBER type")
	return Token(Symbols.NUMBER, int(x_number.value) >> int(y_number.value))

LEL_BIT = {
	"&": _and,
	"|": _or,
	"^": _xor,
	"~": _not,
	"<<": _left_shift,
	">>": _right_shift,
}



import os
from pylel.token import Symbols, Token
from pylel.tools import read_file

def _perform_read_file_sync(scope):
	def func(file_name, encoding):
		if file_name.type == Symbols.STRING and encoding.type == Symbols.STRING:
			from pylel.evaluate.evaluate_expr import find_base_path
			file_name = os.path.join(find_base_path(scope), file_name.value)
			return Token(Symbols.STRING, read_file(file_name, encoding.value))
		else:
			raise Exception("Requires filename and encoding to be strings. Got {} and {}"\
					.format(file_name, encoding))
	return func


def read_file_sync(evaluate_expr, scope, expr):
	if len(expr) < 3:
		raise Exception("Requires filename and encoding.")
	return _perform_read_file_sync(scope)\
	(evaluate_expr(scope, expr[1]), evaluate_expr(scope, expr[2]))
